#include <robot_dart/robot_dart_simu.hpp>

#ifdef GRAPHIC
#include <robot_dart/gui/magnum/graphics.hpp>
#endif

/// namespace alias for ease of use
namespace rd = robot_dart;

int main()
{
    /// Create simulator object
    rd::RobotDARTSimu simu(0.001);

    /// Create graphics
#ifdef GRAPHIC
    auto graphics = std::make_shared<rd::gui::magnum::Graphics>();
    simu.set_graphics(graphics);
    graphics->look_at({0., 2., 1.}, {0., 0., 0.}); // position_of_the_camera, point_to_look_at
#endif

    /// Load robot
    auto robot = std::make_shared<rd::Robot>("arm.urdf");
    /// Fix robot to world frame
    robot->fix_to_world();
    /// Change the actuator type
    robot->set_actuator_types("servo"); // available actuator types: torque, servo, velocity, locked, mimic
    /// Put the robot in some joint positions
    robot->set_positions(rd::make_vector({0., M_PI / 3., M_PI / 4., -M_PI / 2.}));
    // robot->set_positions(rd::make_vector({M_PI/6., M_PI/3., -M_PI/4., M_PI/2.}));
    auto body_names = robot->body_names();
    for (auto& b : body_names)
        std::cout << b << " ";
    std::cout << std::endl;
    robot->set_draw_axis(body_names[0], 0.5);
    robot->set_draw_axis(body_names.back(), 0.25);

    /// Add robot to the world
    simu.add_robot(robot);

    /// Add floor
    simu.add_floor();

    /// Analytical Forward Kinematics
    double L0 = 0.0545;
    double L1 = 0.123;
    double L2 = 0.317;
    double L3 = 0.202;
    double L4 = 0.1605;

    auto joints = robot->positions();

    Eigen::Isometry3d tf_01 = Eigen::Isometry3d::Identity();
    tf_01.linear() = dart::math::eulerZYXToMatrix({joints[0], 0., 0.});
    tf_01.translation() << 0., 0., L0;

    Eigen::Isometry3d tf_12 = Eigen::Isometry3d::Identity();
    tf_12.linear() = dart::math::eulerZYXToMatrix({0., joints[1], 0.});
    tf_12.translation() << 0., 0., L1;

    Eigen::Isometry3d tf_23 = Eigen::Isometry3d::Identity();
    tf_23.linear() = dart::math::eulerZYXToMatrix({0., joints[2], 0.});
    tf_23.translation() << 0., 0., L2;

    Eigen::Isometry3d tf_34 = Eigen::Isometry3d::Identity();
    tf_34.linear() = dart::math::eulerZYXToMatrix({0., joints[3], 0.});
    tf_34.translation() << 0., 0., L3;

    Eigen::Isometry3d tf_45 = Eigen::Isometry3d::Identity();
    // tf_45.linear() = dart::math::eulerZYXToMatrix({0., 0., 0.});
    tf_45.translation() << 0., 0., L4;

    Eigen::Isometry3d tf = tf_01 * tf_12 * tf_23 * tf_34 * tf_45;
    std::cout << "Analytical:" << std::endl;
    std::cout << tf.matrix() << std::endl;

    /// robot_dart Forward Kinematics
    std::cout << "robot_dart:" << std::endl;
    std::cout << robot->body_pose("arm_link_5").matrix() << std::endl;

    /// Run simulation
    while (true) {
        if (simu.step_world())
            break;
    }

    return 0;
}
