import numpy as np
import RobotDART as rd
import dartpy  # OSX breaks if this is imported before RobotDART
import copy
from utils import damped_pseudoinverse, AdT, angle_wrap_multi

# Load a robot
robot = rd.Iiwa()
robot.set_actuator_types("servo")

# Load ghost robot for visualization
robot_ghost = robot.clone_ghost()

# set initial joint positions
target_positions = copy.copy(robot.positions())
target_positions[0] = -2.
target_positions[3] = -np.pi / 2.0
target_positions[5] = np.pi / 2.0
# target_positions[1] = -np.pi / 2.0
robot.set_positions(target_positions)

# get end-effector pose
eef_link_name = "iiwa_link_ee"
tf_desired = robot.body_pose(eef_link_name)
print(tf_desired)

# set robot back to zero positions
robot.reset()

# function to compute error in Transformation space
def error(tf, tf_desired):
    return rd.math.logMap(tf.inverse().multiply(tf_desired))

# optimization Newton-Raphson
def ik_jac(init_positions, tf_desired, step = np.pi/4., max_iter = 100, min_error = 1e-6):
    pos = init_positions
    for _ in range(max_iter):
        robot.set_positions(pos)
        tf = robot.body_pose(eef_link_name)
        Ad_tf = AdT(tf)
        error_in_body_frame = error(tf, tf_desired)
        error_in_world_frame = Ad_tf @ error_in_body_frame

        ferror = np.linalg.norm(error_in_world_frame)
        if ferror < min_error:
            break

        jac = robot.jacobian(eef_link_name) # this is in world frame
        jac_pinv = damped_pseudoinverse(jac) # np.linalg.pinv(jac) # get pseudo-inverse

        delta_pos = jac_pinv @ error_in_world_frame
        # We can limit the delta_pos to avoid big steps due to pseudo-inverse instability
        # you can play with the step value to see the effect
        for i in range(delta_pos.shape[0]):
            if delta_pos[i] > step:
                delta_pos[i] = step
            elif delta_pos[i] < -step:
                delta_pos[i] = -step

        pos = pos + delta_pos

    # We would like to wrap the final joint positions to [-pi,pi)
    pos = angle_wrap_multi(pos)
    print('Final error:', ferror)

    return pos

tt = copy.copy(robot.positions())
res = ik_jac(tt, tf_desired)
print('IK result:', res)
print('Target joint:', target_positions)

# Create simulator object
simu = rd.RobotDARTSimu()

# Create Graphics
gconfig = rd.gui.GraphicsConfiguration(1024, 768) # Create a window of 1024x768 resolution/size
graphics = rd.gui.Graphics(gconfig) # create graphics object with configuration
simu.set_graphics(graphics)
graphics.look_at([3., 1., 2.], [0., 0., 0.])

# Add robot and nice floor
simu.add_robot(robot)
robot_ghost.set_positions(target_positions)
simu.add_robot(robot_ghost)
simu.add_checkerboard_floor()

robot.set_positions(res)

# Run simulation
while True:
    if simu.step_world():
        break
